from __future__ import absolute_import, division

import logging
import time

import numpy as np
from scipy.sparse import spdiags
from scipy.sparse.linalg import LinearOperator, aslinearoperator, lsqr

logger = logging.getLogger(__name__)

# Size of info vector in case of infinite iterations
_allocSize = 10000

# Machine epsilon
_eps = np.spacing(1)

# Exit conditions (constants).
EXIT_ROOT_FOUND = 1
EXIT_BPSOL_FOUND = 2
EXIT_LEAST_SQUARES = 3
EXIT_OPTIMAL = 4
EXIT_ITERATIONS = 5
EXIT_LINE_ERROR = 6
EXIT_SUBOPTIMAL_BP = 7
EXIT_MATVEC_LIMIT = 8
EXIT_ACTIVE_SET = 9
EXIT_CONVERGED_spgline = 0
EXIT_ITERATIONS_spgline = 1
EXIT_NODESCENT_spgline = 2


# private classes
class _LSQRprod(LinearOperator):
    """LSQR operator.

    This operator is used to augument the spgl1 operator during subspace
    minimization via LSQR.
    """

    def __init__(self, A, nnz_idx, ebar, n):
        self.A = A
        self.nnz_idd = nnz_idx
        self.ebar = ebar
        self.nbar = np.size(ebar)
        self.n = n
        self.shape = (A.shape[0], self.nbar)
        self.dtype = A.dtype

    def _matvec(self, x):
        y = np.zeros(self.n, dtype=x.dtype)
        y[self.nnz_idd] = x - (1.0 / self.nbar) * np.dot(
            np.dot(np.conj(self.ebar), x), self.ebar
        )
        z = self.A.matvec(y)
        return z

    def _rmatvec(self, x):
        y = self.A.rmatvec(x)
        z = y[self.nnz_idd] - (1.0 / self.nbar) * np.dot(
            np.dot(np.conj(self.ebar), y[self.nnz_idd]), self.ebar
        )
        return z


class _blockdiag(LinearOperator):
    """Block-diagonal operator.

    This operator is created to work with the spg_mmv solver, which solves
    a multi-measurement basis pursuit denoise problem. Model and data are
    matrices of size ``N x G`` and ``M x G`` respectively,
    and the operator ``A`` is applied to each column of the vectors.
    """

    def __init__(self, A, m, n, g):
        self.m = m
        self.n = n
        self.g = g
        self.A = A
        self.AH = A.H
        self.shape = (m * g, n * g)
        self.dtype = A.dtype

    def _matvec(self, x):
        x = x.reshape(self.n, self.g)
        y = self.A.matmat(x)
        return y.ravel()

    def _rmatvec(self, x):
        x = x.reshape(self.m, self.g)
        y = self.AH.matmat(x)
        return y.ravel()


# private methods
def _printf(fid, message):
    """Print a message in file (fid=file ID) or on screen (fid=None)"""
    if fid is None:
        print(message)
    else:
        fid.write(message)


def _oneprojector_i(b, tau):
    n = b.size
    x = np.zeros(n, dtype=b.dtype)
    bNorm = np.linalg.norm(b, 1)

    if tau >= bNorm:
        return b.copy()
    elif tau < np.spacing(1):
        pass
    else:
        idx = np.argsort(b)[::-1]
        b = b[idx]

        csb = np.cumsum(b) - tau
        alpha = np.zeros(n + 1)
        alpha[1:] = csb / (np.arange(n) + 1.0)
        alphaindex = np.where(alpha[1:] >= b)[0]
        if alphaindex.any():
            alphaPrev = alpha[alphaindex[0]]
        else:
            alphaPrev = alpha[-1]

        x[idx] = b - alphaPrev
        x[x < 0] = 0
    return x


def _oneprojector_d(b, d, tau):
    n = b.size
    x = np.zeros(n, dtype=b.dtype)

    if tau >= np.linalg.norm(d * b, 1):
        x = b.copy()
    elif tau < np.spacing(1):
        pass
    else:
        # Preprocessing
        idx = np.argsort(b / d)[::-1]
        b = b[idx]
        d = d[idx]

        # Optimize
        csdb = np.cumsum(d * b)
        csd2 = np.cumsum(d * d)
        alpha1 = (csdb - tau) / csd2
        alpha2 = b / d
        ggg = np.where(alpha1 >= alpha2)[0]
        if ggg.size == 0:
            i = n
        else:
            i = ggg[0]
        if i > 0:
            soft = alpha1[i - 1]
            x[idx[0:i]] = b[0:i] - d[0:i] * max(0, soft)
    return x


def _oneprojector_di(b, d, tau):
    if np.isscalar(d):
        p = _oneprojector_i(b, tau / abs(d))
    else:
        p = _oneprojector_d(b, d, tau)
    return p


def oneprojector(b, d, tau):
    """One projector.

    Projects b onto the (weighted) one-norm ball of radius tau.
    If d=1 solves the problem::

        minimize_x  ||b-x||_2  subject to  ||x||_1 <= tau.

    else::

        minimize_x  ||b-x||_2  subject to  ||Dx||_1 <= tau.

    Parameters
    ----------
    b : ndarray
        Input vector to be projected.
    d : {ndarray, float}
        Weight vector (or scalar)
    tau : float
        Radius of one-norm ball.

    Returns
    -------
    x : array_like
        Projected vector

    """
    if not np.isscalar(d) and b.size != d.size:
        raise ValueError("vectors b and d must have the same length")

    if np.isscalar(d) and d == 0:
        x = b.copy()
    else:
        # Get sign of b and set to absolute values
        s = np.sign(b)
        b = np.abs(b)

        # Perform the projection
        if np.isscalar(d):
            x = _oneprojector_di(b, 1.0, tau / d)
        else:
            d = np.abs(d)
            # Get index of non-zero entries of d, set x equal b for others
            idx = np.where(d > np.spacing(1))
            x = b.copy()
            x[idx] = _oneprojector_di(b[idx], d[idx], tau)
        # Restore signs in x
        x *= s.astype(x.dtype)

    return x


def _norm_l1_primal(x, weights):
    """L1 norm with weighted input vector

    Parameters
    ----------
    x : ndarray
        Input array
    weights : {float, ndarray}
        Weights

    Returns
    -------
    . : float
        L1 norm

    """
    return np.linalg.norm(x * weights, 1)


def _norm_l1_dual(x, weights):
    """L_inf norm with weighted input vector (dual to L1 norm)

    Parameters
    ----------
    x : ndarray
        Input array
    weights : {float, ndarray}
        Weights

    Returns
    -------
    . : float
        L_inf norm

    """
    return np.linalg.norm(x / weights, np.inf)


def _norm_l1_project(x, weights, tau):
    """Projection onto the one-norm ball

    Parameters
    ----------
    x : ndarray
        Input array
    weights : {float, ndarray}, optional
        Weights
    tau : float
        Projection radius

    Returns
    -------
    xproj : float
        Projected array

    """
    if not np.iscomplexobj(x):
        xproj = oneprojector(x, weights, tau)
    else:
        xa = np.abs(x)
        idx = xa < _eps
        xc = oneprojector(xa, weights, tau)
        xc /= xa + 1e-10
        xc[idx] = 0
        xproj = x * xc
    return xproj


def _norm_l12_primal(g, x, weights):
    """L1 norm with weighted input vector with number of groups equal to g
    Parameters
    ----------
    g : int
        Number of groups
    x : ndarray
        Input array
    weights : {float, ndarray}, optional
        Weights
    Returns
    -------
    nrm : float
        Group norm
    """
    m = x.size // g
    xx = x.copy()
    xx = xx.reshape(m, g)
    if np.iscomplexobj(xx):
        xx = np.abs(xx)
    nrm = np.sum(weights * np.sqrt(np.sum(xx**2, axis=1)))
    return nrm


def _norm_l12_dual(g, x, weights):
    """L_inf norm with weighted input vector with number of groups equal to g
    Parameters
    ----------
    g : int
        Number of groups
    x : ndarray
        Input array
    weights : {float, ndarray}, optional
        Weights
    Returns
    -------
    nrm : float
        Group norm
    """
    m = x.size // g
    xx = x.copy()
    xx = xx.reshape(m, g)
    if np.iscomplexobj(xx):
        xx = np.abs(xx)
    nrm = np.linalg.norm(np.sqrt(np.sum(xx**2, axis=1)) / weights, np.inf)
    return nrm


def _norm_l12_project(g, x, weights, tau):
    """Projection with number of groups equal to g
    Parameters
    ----------
    g : int
        Number of groups
    x : ndarray
        Input array
    weights : {float, ndarray}, optional
        Weights
    tau : float
        Projection radius
    Returns
    -------
    x : float
        Projected array
    """
    m = x.size // g
    xx = x.copy()
    xx = xx.reshape(m, g)
    if np.iscomplexobj(xx):
        xx = np.abs(xx)
    xa = np.sqrt(np.sum(xx**2, axis=1))

    idx = xa < np.spacing(1)
    xc = oneprojector(xa, weights, tau)
    xc = xc / xa
    xc[idx] = 0
    xx = spdiags(xc, 0, m, m) * xx
    return xx.flatten()


def norm_l1nn_primal(x, weights):
    """Non-negative L1 gauge function

    Parameters
    ----------
    x : ndarray
        Input array
    weights : {float, ndarray}, optional
        Weights

    Returns
    -------
    p : float
        Norm

    """
    if np.any(x < 0):
        p = np.inf
    else:
        p = np.linalg.norm(x * weights, 1)
    return p


def norm_l1nn_dual(x, weights):
    """Dual of non-negative L1 gauge function

    Parameters
    ----------
    x : ndarray
        Input array
    weights : {float, ndarray}, optional
        Weights

    Returns
    -------
    p : float
        Norm

    """
    xx = x.copy()
    xx[xx < 0] = 0
    p = np.linalg.norm(xx / weights, np.inf)
    return p


def norm_l1nn_project(x, weights, tau):
    """Projection onto the non-negative part of the one-norm ball

    Parameters
    ----------
    x : ndarray
        Input array
    weights : {float, ndarray}, optional
        Weights
    tau : float
        Projection radius

    Returns
    -------
    . : float
        Projected array

    """
    xx = x.copy()
    xx[xx < 0] = 0
    return _norm_l1_project(xx, weights, tau)


def norm_l12nn_primal(g, x, weights):
    """Non-negative L1 norm with weighted input vector with number
    of groups equal to g

    Parameters
    ----------
    g : int
        Number of groups
    x : ndarray
        Input array
    weights : {float, ndarray}, optional
        Weights

    Returns
    -------
    nrm : float
        Group norm

    """
    m = x.size // g

    if np.any(x < 0):
        nrm = np.inf
    else:
        xm = x.reshape(m, g)
        if np.iscomplexobj(x):
            xm = np.abs(xm)
        nrm = np.sum(weights * np.sqrt(np.sum(xm**2, axis=1)))
    return nrm


def norm_l12nn_dual(g, x, weights):
    """Dual on non-legative L1L norm with weighted input vector
    with number of groups equal to g

    Parameters
    ----------
    g : int
        Number of groups
    x : ndarray
        Input array
    weights : {float, ndarray}, optional
        Weights

    Returns
    -------
    nrm : float
        Group norm

    """
    m = x.size // g
    xx = x.copy()
    xx = xx.reshape(m, g)

    if np.iscomplexobj(xx):
        xx = np.abs(xx)
    xx[xx < 0] = 0
    nrm = np.linalg.norm(np.sqrt(np.sum(xx**2, axis=1)) / weights, np.inf)
    return nrm


def norm_l12nn_project(g, x, weights, tau):
    """Projection with number of groups equal to g

    Parameters
    ----------
    g : int
        Number of groups
    x : ndarray
        Input array
    weights : {float, ndarray}, optional
        Weights
    tau : float
        Projection radius

    Returns
    -------
    x : float
        Projected array

    """
    xx = x.copy()
    if np.iscomplexobj(xx):
        xx = np.abs(xx)
    xx[xx < 0] = 0
    return _norm_l12_project(g, xx, weights, tau)


def _spg_line_curvy(x, g, fmax, A, b, project, weights, tau):
    """Projected backtracking linesearch.

    On entry, g is the (possibly scaled) steepest descent direction.

    Parameters
    ----------
    x : ndarray
        Input array
    g : ndarray
        Input gradient
    fmax : float
        Maximum residual norm
    A : {sparse matrix, ndarray, LinearOperator}
        Operator
    b : ndarray
        Data
    project : func, optional
        Projection function
    weights : {float, ndarray}, optional
        Weights ``W`` in ``||Wx||_1``
    tau : float, optional
        Projection radium

    Returns
    -------
    fnew : float
        Residual norm after linesearch projection
    xnew : ndarray
        Model after linesearch projection
    rnew : ndarray
        Residual after linesearch projection
    niters : int
        Number of iterations
    step : int
        Final step
    err : int
        Error flag
    timeproject : float
        Time in secs for projection
    timematprod : float
        Time in secs for matvec computations

    """
    gamma = 1e-4
    maxiters = 10
    step = 1.0
    snorm = 0.0
    scale = 1.0
    nsafe = 0
    niters = 0
    n = x.size
    timeproject = 0
    timematprod = 0
    while 1:
        # Evaluate trial point and function value.
        start_time_project = time.time()
        xnew = project(x - (step * scale * g).astype(g.dtype), weights, tau)
        timeproject += time.time() - start_time_project
        start_time_matprod = time.time()
        rnew = b - A.matvec(xnew)
        timematprod += time.time() - start_time_matprod
        fnew = np.abs(np.conj(rnew).dot(rnew)) / 2.0
        s = xnew - x
        gts = scale * np.real(np.dot(np.conj(g), s))

        if gts >= 0:
            err = EXIT_NODESCENT_spgline
            break

        if fnew < fmax + gamma * step * gts:
            err = EXIT_CONVERGED_spgline
            break
        elif niters >= maxiters:
            err = EXIT_ITERATIONS_spgline
            break

        # New linesearch iteration.
        niters += 1
        step /= 2.0

        # Safeguard: If stepMax is huge, then even damped search
        # directions can give exactly the same point after projection. If
        # we observe this in adjacent iterations, we drastically damp the
        # next search direction.
        snormold = snorm
        snorm = np.linalg.norm(s) / np.sqrt(n)
        if abs(snorm - snormold) <= 1e-6 * snorm:
            gnorm = np.linalg.norm(g) / np.sqrt(n)
            scale = snorm / gnorm / (2.0**nsafe)
            nsafe += 1.0
    return fnew, xnew, rnew, niters, step, err, timeproject, timematprod


def _spg_line(f, x, d, gtd, fmax, A, b):
    """Non-monotone linesearch.

    Parameters
    ----------
    f : float
        Residual norm
    x : ndarray
        Input array
    d : float
        Difference between input array and proposed projected array
    gtd : float
        Dot product between gradient and d
    fmax : float
        Maximum residual norm
    A : {sparse matrix, ndarray, LinearOperator}
        Operator
    b : ndarray
        Data

    Returns
    -------
    fnew : float
        Residual norm after linesearch projection
    xnew : ndarray
        Model after linesearch projection
    xnew : ndarray
        Residual after linesearch projection
    niters : int
        Number of iterations
    err : int
        Error flag
    timematprod : float
        Time in secs for matvec computations

    """
    maxiters = 10
    step = 1.0
    niters = 0
    gamma = 1e-4
    gtd = -abs(gtd)
    timematprod = 0
    while 1:
        # Evaluate trial point and function value.
        xnew = x + step * d
        start_time_matprod = time.time()
        rnew = b - A.matvec(xnew)
        timematprod += time.time() - start_time_matprod
        fnew = abs(np.conj(rnew).dot(rnew)) / 2.0

        # Check exit conditions.
        if fnew < fmax + gamma * step * gtd:  # Sufficient descent condition.
            err = EXIT_CONVERGED_spgline
            break
        elif niters >= maxiters:  # Too many linesearch iterations.
            err = EXIT_ITERATIONS_spgline
            break

        # New line-search iteration.
        niters += 1

        # Safeguarded quadratic interpolation.
        if step <= 0.1:
            step /= 2.0
        else:
            tmp = (-gtd * step**2.0) / (2 * (fnew - f - step * gtd))
            if tmp < 0.1 or tmp > 0.9 * step or np.isnan(tmp):
                tmp = step / 2.0
            step = tmp
    return fnew, xnew, rnew, niters, err, timematprod


def _active_vars(x, g, nnz_idx, opttol, weights, dual_norm):
    """Find the current active set.

    Returns
    -------
    nnz_x : int
        Number of nonzero elements in x.
    nnz_g : int
        Number of elements in nnz_idx.
    nnz_idx : array_like
        Vector of primal/dual indicators.
    nnz_diff : int
        Number of elements that changed in the support.

    """
    xtol = min([0.1, 10.0 * opttol])
    gtol = min([0.1, 10.0 * opttol])
    gnorm = dual_norm(g, weights)
    nnz_old = nnz_idx

    # Reduced costs for positive and negative parts of x.
    z1 = gnorm + g
    z2 = gnorm - g

    # Primal/dual based indicators.
    xpos = (x > xtol) & (z1 < gtol)
    xneg = (x < -xtol) & (z2 < gtol)
    nnz_idx = xpos | xneg

    # Count is based on simple primal indicator.
    nnz_x = np.sum(np.abs(x) >= xtol)
    nnz_g = np.sum(nnz_idx)

    if nnz_old is None:
        nnz_diff = np.inf
    else:
        nnz_diff = np.sum(nnz_idx != nnz_old)

    return nnz_x, nnz_g, nnz_idx, nnz_diff


def spgl1(
    A,
    b,
    tau=0,
    sigma=0,
    x0=None,
    fid=None,
    verbosity=0,
    iter_lim=None,
    n_prev_vals=3,
    bp_tol=1e-6,
    ls_tol=1e-6,
    opt_tol=1e-4,
    dec_tol=1e-4,
    step_min=1e-16,
    step_max=1e5,
    active_set_niters=np.inf,
    subspace_min=False,
    iscomplex=False,
    max_matvec=np.inf,
    weights=None,
    project=_norm_l1_project,
    primal_norm=_norm_l1_primal,
    dual_norm=_norm_l1_dual,
):
    r"""SPGL1 solver.

    Solve basis pursuit (BP), basis pursuit denoise (BPDN), or LASSO problems
    [1]_ [2]_ depending on the choice of ``tau`` and ``sigma``::

        (BP)     minimize  ||x||_1  subj. to  Ax = b

        (BPDN)   minimize  ||x||_1  subj. to  ||Ax-b||_2 <= sigma

        (LASSO)  minimize  ||Ax-b||_2  subj, to  ||x||_1 <= tau

    The matrix ``A`` may be square or rectangular (over-determined or
    under-determined), and may have any rank.

    Parameters
    ----------
    A : {sparse matrix, ndarray, LinearOperator}
        Representation of an m-by-n matrix.  It is required that
        the linear operator can produce ``Ax`` and ``A^T x``.
    b : array_like, shape (m,)
        Right-hand side vector ``b``.
    tau : float, optional
        LASSO threshold. If different from ``None``, spgl1 solves LASSO problem
    sigma : float, optional
        BPDN threshold. If different from ``None``, spgl1 solves BPDN problem
    x0 : array_like, shape (n,), optional
        Initial guess of x, if None zeros are used.
    fid : file, optional
        File ID to direct log output, if None print on screen.
    verbosity : int, optional
        0=quiet, 1=some output, 2=more output.
    iter_lim : int, optional
        Max. number of iterations (default if ``10*m``).
    n_prev_vals : int, optional
         Line-search history lenght.
    bp_tol : float, optional
        Tolerance for identifying a basis pursuit solution.
    ls_tol : float, optional
         Tolerance for least-squares solution. Iterations are stopped when the
         ratio between the dual norm of the gradient and the L2 norm of the
         residual becomes smaller or equal to ``ls_tol``.
    opt_tol : float, optional
        Optimality tolerance. More specifically, when using basis pursuit
        denoise, the optimility condition is met when the absolute difference
        between the L2 norm of the residual and the ``sigma`` is smaller than
        ``opt_tol``.
    dec_tol : float, optional
        Required relative change in primal objective for Newton.
        Larger ``decTol`` means more frequent Newton updates.
    step_min : float, optional
        Minimum spectral step.
    step_max : float, optional
        Maximum spectral step.
    active_set_niters : float, optional
        Maximum number of iterations where no change in support is tolerated.
        Exit with EXIT_ACTIVE_SET if no change is observed for ``activeSetIt``
        iterations
    subspace_min : bool, optional
        Subspace minimization (``True``) or not (``False``)
    iscomplex : bool, optional
        Problem with complex variables (``True``) or not (``False``)
    max_matvec : int, optional
        Maximum matrix-vector multiplies allowed
    weights : {float, ndarray}, optional
        Weights ``W`` in ``||Wx||_1``
    project : func, optional
        Projection function
    primal_norm : func, optional
        Primal norm evaluation fun
    dual_norm : func, optional
         Dual norm eval function

    Returns
    -------
    x : array_like, shape (n,)
        Inverted model
    r : array_like, shape (m,)
        Final residual
    g : array_like, shape (h,)
        Final gradient
    info : dict
        Dictionary with the following information:

        ``tau``, final value of tau (see sigma above)

        ``rnorm``, two-norm of the optimal residual

        ``rgap``, relative duality gap (an optimality measure)

        ``gnorm``, Lagrange multiplier of (LASSO)

        ``stat``,
           ``1``: found a BPDN solution,
           ``2``: found a BP solution; exit based on small gradient,
           ``3``: found a BP solution; exit based on small residual,
           ``4``: found a LASSO solution,
           ``5``: error: too many iterations,
           ``6``: error: linesearch failed,
           ``7``: error: found suboptimal BP solution,
           ``8``: error: too many matrix-vector products

        ``niters``, number of iterations

        ``nProdA``, number of multiplications with A

        ``nProdAt``, number of multiplications with A'

        ``n_newton``, number of Newton steps

        ``time_project``, projection time (seconds)

        ``time_matprod``, matrix-vector multiplications time (seconds)

        ``time_total``, total solution time (seconds)

        ``niters_lsqr``, number of lsqr iterations (if ``subspace_min=True``)

        ``xnorm1``, L1-norm model solution history through iterations

        ``rnorm2``, L2-norm residual history through iterations

        ``lambdaa``, Lagrange multiplier history through iterations

    References
    ----------
    .. [1] E. van den Berg and M. P. Friedlander, "Probing the Pareto frontier
             for basis pursuit solutions", SIAM J. on Scientific Computing,
             31(2):890-912. (2008).
    .. [2] E. van den Berg and M. P. Friedlander, "Sparse optimization with
             least-squares constraints", Tech. Rep. TR-2010-02, Dept of
             Computer Science, Univ of British Columbia (2010).

    """
    start_time = time.time()

    A = aslinearoperator(A)
    m, n = A.shape

    if tau == 0:
        single_tau = False
    else:
        single_tau = True

    if iter_lim is None:
        iter_lim = 10 * m

    max_line_errors = 10  # Maximum number of line-search failures.
    piv_tol = 1e-12  # Threshold for significant Newton step.
    max_matvec = max(3, max_matvec)  # Max number of allowed matvec/rmatvec.

    # Initialize local variables.
    niters = 0  # Total SPGL1 iterations.
    niters_lsqr = 0  # Total LSQR iterations.
    nprodA = 0  # Number of matvec operations
    nprodAt = 0  # Number of rmatvec operations
    last_fv = np.full(10, -np.inf)  # Last m function values.
    nline_tot = 0  # Total number of linesearch steps.
    print_tau = False
    n_newton = 0  # Number of Newton iterations
    bnorm = np.linalg.norm(b)
    stat = False
    time_project = 0  # Time spent in projections
    time_matprod = 0  # Time spent in matvec computations
    nnz_niters = 0  # No. of iterations with fixed pattern.
    nnz_idx = None  # Active-set indicator.
    subspace = False  # Flag if did subspace min in current itn.
    stepg = 1  # Step length for projected gradient.
    test_updatetau = False  # Previous step did not update tau

    # Determine initial x and see if problem is complex
    realx = np.isreal(A).all() and np.isreal(b).all()
    if x0 is None:
        x = np.zeros(n, dtype=b.dtype)
    else:
        x = np.asarray(x0)

    # Override realx when iscomplex flag is set
    if iscomplex:
        realx = False

    # Check if all weights (if any) are strictly positive. In previous
    # versions we also checked if the number of weights was equal to
    # n. In the case of multiple measurement vectors, this no longer
    # needs to apply, so the check was removed.
    if weights is not None:
        if not np.isfinite(weights).all():
            raise ValueError("Entries in weights must be finite")
        if np.any(weights <= 0):
            raise ValueError("Entries in weights must be strictly positive")
    else:
        weights = 1

    # Quick exit if sigma >= ||b||.  Set tau = 0 to short-circuit the loop.
    if bnorm <= sigma:
        logger.warning("W: sigma >= ||b||.  Exact solution is x = 0.")
        tau = 0
        single_tau = True

    # Do not do subspace minimization if x is complex.
    if not realx and subspace_min:
        logger.warning("W: Subspace minimization disabled when variables are complex.")
        subspace_min = False

    #% Pre-allocate iteration info vectors
    xnorm1 = np.zeros(min(iter_lim + 1, _allocSize))
    rnorm2 = np.zeros(min(iter_lim + 1, _allocSize))
    lambdaa = np.zeros(min(iter_lim + 1, _allocSize))

    # Log header.
    if verbosity >= 1:
        _printf(fid, "")
        _printf(fid, "=" * 80 + "")
        _printf(fid, "SPGL1")
        _printf(fid, "=" * 80 + "")
        _printf(fid, "%-22s: %8i %4s" % ("No. rows", m, ""))
        _printf(fid, "%-22s: %8i\n" % ("No. columns", n))
        _printf(fid, "%-22s: %8.2e %4s" % ("Initial tau", tau, ""))
        _printf(fid, "%-22s: %8.2e\n" % ("Two-norm of b", bnorm))
        _printf(fid, "%-22s: %8.2e %4s" % ("Optimality tol", opt_tol, ""))
        if single_tau:
            _printf(fid, "%-22s: %8.2e\n" % ("Target one-norm of x", tau))
        else:
            _printf(fid, "%-22s: %8.2e\n" % ("Target objective", sigma))
        _printf(fid, "%-22s: %8.2e %4s" % ("Basis pursuit tol", bp_tol, ""))
        _printf(fid, "%-22s: %8i\n" % ("Maximum iterations", iter_lim))
        if verbosity >= 2:
            if single_tau:
                logb = "%5i  %13.7e  %13.7e  %9.2e  %6.1f  %6i  %6i %6s"
                logh = "%5s  %13s  %13s  %9s  %6s  %6s  %6s\n"
                _printf(
                    fid,
                    logh
                    % (
                        "iterr",
                        "Objective",
                        "Relative Gap",
                        "gnorm",
                        "stepg",
                        "nnz_x",
                        "nnz_g",
                    ),
                )
            else:
                logb = "%5i  %13.7e  %13.7e  %9.2e  %9.3e  %6.1f  %6i  %6i %6s"
                logh = "%5s  %13s  %13s  %9s  %9s  %6s  %6s  %6s  %6s\n"
                _printf(
                    fid,
                    logh
                    % (
                        "iterr",
                        "Objective",
                        "Relative Gap",
                        "Rel Error",
                        "gnorm",
                        "stepg",
                        "nnz_x",
                        "nnz_g",
                        "tau",
                    ),
                )

    # Project the starting point and evaluate function and gradient.
    start_time_project = time.time()
    x = project(x, weights, tau)
    time_project += time.time() - start_time_project
    start_time_matvec = time.time()
    r = b - A.matvec(x)  # r = b - Ax
    g = -A.rmatvec(r)  # g = -A'r
    time_matprod += time.time() - start_time_matvec
    f = np.linalg.norm(r) ** 2 / 2.0
    nprodA += 1
    nprodAt += 1

    # Required for nonmonotone strategy.
    last_fv[0] = f
    fbest = f
    xbest = x.copy()
    fold = f

    # Compute projected gradient direction and initial step length.
    start_time_project = time.time()
    dx = project(x - g, weights, tau) - x
    time_project += time.time() - start_time_project
    dxnorm = np.linalg.norm(dx, np.inf)
    if dxnorm < (1.0 / step_max):
        gstep = step_max
    else:
        gstep = min(step_max, max(step_min, 1.0 / dxnorm))

    # Main iteration loop.
    while 1:
        # Test exit conditions.

        # Compute quantities needed for log and exit conditions.
        gnorm = dual_norm(-g, weights)
        rnorm = np.linalg.norm(r)
        gap = np.dot(np.conj(r), r - b) + tau * gnorm
        rgap = abs(gap) / max(1.0, f)
        aerror1 = rnorm - sigma
        aerror2 = f - sigma**2.0 / 2.0
        rerror1 = abs(aerror1) / max(1.0, rnorm)
        rerror2 = abs(aerror2) / max(1.0, f)

        #% Count number of consecutive iterations with identical support.
        nnz_old = nnz_idx
        nnz_x, nnz_g, nnz_idx, nnz_diff = _active_vars(
            x, g, nnz_idx, opt_tol, weights, dual_norm
        )
        if nnz_diff:
            nnz_niters = 0
        else:
            nnz_niters += nnz_niters
            if nnz_niters + 1 >= active_set_niters:
                stat = EXIT_ACTIVE_SET

        # Single tau: Check if were optimal.
        # The 2nd condition is there to guard against large tau.
        if single_tau:
            if rgap <= opt_tol or rnorm < opt_tol * bnorm:
                stat = EXIT_OPTIMAL
        else:  # Multiple tau: Check if found root and/or if tau needs updating.
            # Test if a least-squares solution has been found
            if gnorm <= ls_tol * rnorm:
                stat = EXIT_LEAST_SQUARES
            if rgap <= max(opt_tol, rerror2) or rerror1 <= opt_tol:
                # The problem is nearly optimal for the current tau.
                # Check optimality of the current root.
                if rnorm <= sigma:
                    stat = EXIT_SUBOPTIMAL_BP  # Found suboptimal BP sol.
                if rerror1 <= opt_tol:
                    stat = EXIT_ROOT_FOUND  # Found approx root.
                if rnorm <= bp_tol * bnorm:
                    stat = EXIT_BPSOL_FOUND  # Resid minimzd -> BP sol.
            fchange = np.abs(f - fold)
            test_relchange1 = fchange <= dec_tol * f
            test_relcchange2 = fchange <= 1e-1 * f * (np.abs(rnorm - sigma))
            test_updatetau = (
                (
                    (test_relchange1 and rnorm > 2 * sigma)
                    or (test_relcchange2 and rnorm <= 2 * sigma)
                )
                and not stat
                and not test_updatetau
            )

            if test_updatetau:
                # Update tau.
                tau_old = tau
                tau = max(0, tau + (rnorm * aerror1) / gnorm)
                n_newton += 1
                print_tau = np.abs(tau_old - tau) >= 1e-6 * tau  # For log only.
                if tau < tau_old:
                    # The one-norm ball has decreased. Need to make sure that
                    # the next iterate is feasible, which we do by projecting it.
                    start_time_project = time.time()
                    x = project(x, weights, tau)
                    time_project += time.time() - start_time_project

                    # Update the residual, gradient, and function value
                    start_time_matvec = time.time()
                    r = b - A.matvec(x)
                    g = -A.rmatvec(r)
                    time_matprod += time.time() - start_time_matvec

                    f = np.linalg.norm(r) ** 2 / 2.0
                    nprodA += 1
                    nprodAt += 1

                    # Reset the function value history.
                    last_fv = np.full(10, -np.inf)
                    last_fv[1] = f

        # Too many iterations and not converged.
        if not stat and niters >= iter_lim:
            stat = EXIT_ITERATIONS

        # Print log, update history and act on exit conditions.
        if verbosity >= 2 and (
            ((niters < 10) or (iter_lim - niters < 10) or (niters % 10 == 0))
            or single_tau
            or print_tau
            or stat
        ):
            tauflag = "              "
            subflag = ""
            if print_tau:
                tauflag = " %13.7e" % tau
            if subspace:
                subflag = " S %2i" % niters_lsqr
            if single_tau:
                _printf(
                    fid,
                    logb
                    % (
                        niters,
                        rnorm,
                        rgap,
                        gnorm,
                        np.log10(stepg),
                        nnz_x,
                        nnz_g,
                        subflag,
                    ),
                )
                if subspace:
                    _printf(fid, "  %s" % subflag)
            else:
                _printf(
                    fid,
                    logb
                    % (
                        niters,
                        rnorm,
                        rgap,
                        rerror1,
                        gnorm,
                        np.log10(stepg),
                        nnz_x,
                        nnz_g,
                        tauflag + subflag,
                    ),
                )
        print_tau = False
        subspace = False

        # Update history info
        if niters > 0 and niters % _allocSize == 0:  # enlarge allocation
            allocincrement = min(_allocSize, iter_lim - xnorm1.shape[0])
            xnorm1 = np.hstack((xnorm1, np.zeros(allocincrement)))
            rnorm2 = np.hstack((rnorm2, np.zeros(allocincrement)))
            lambdaa = np.hstack((lambdaa, np.zeros(allocincrement)))

        xnorm1[niters] = primal_norm(x, weights)
        rnorm2[niters] = rnorm
        lambdaa[niters] = gnorm

        if stat:
            break

        # Iterations begin here.
        niters += 1
        xold = x.copy()
        fold = f.copy()
        gold = g.copy()
        rold = r.copy()

        while 1:
            # Projected gradient step and linesearch.
            (
                f,
                x,
                r,
                niter_line,
                stepg,
                lnerr,
                time_project_curvy,
                time_matprod_curvy,
            ) = _spg_line_curvy(x, gstep * g, max(last_fv), A, b, project, weights, tau)
            time_project += time_project_curvy
            time_matprod += time_matprod_curvy
            nprodA += niter_line + 1
            nline_tot += niter_line
            if nprodA + nprodAt > max_matvec:
                stat = EXIT_MATVEC_LIMIT
                break

            if lnerr:
                # Projected backtrack failed.
                # Retry with feasible dirn linesearch.
                x = xold.copy()
                f = fold
                start_time_project = time.time()
                dx = project(x - gstep * g, weights, tau) - x
                time_project += time.time() - start_time_project
                gtd = np.dot(np.conj(g), dx)
                f, x, r, niter_line, lnerr, time_matprod = _spg_line(
                    f, x, dx, gtd, max(last_fv), A, b
                )
                time_matprod += time_matprod
                nprodA += niter_line + 1
                nline_tot += niter_line
                if nprodA + nprodAt > max_matvec:
                    stat = EXIT_MATVEC_LIMIT
                    break

                if lnerr:
                    # Failed again.
                    # Revert to previous iterates and damp max BB step.
                    x = xold
                    f = fold
                    if max_line_errors <= 0:
                        stat = EXIT_LINE_ERROR
                    else:
                        step_max = step_max / 10.0
                        logger.warning(
                            "Linesearch failed with error %s. "
                            "Damping max BB scaling to %s",
                            lnerr,
                            step_max,
                        )
                        max_line_errors -= 1

            # Subspace minimization (only if active-set change is small).
            if subspace_min:
                start_time_matvec = time.time()
                g = -A.rmatvec(r)
                time_matprod += time.time() - start_time_matvec
                nprodAt += 1
                nnz_x, nnz_g, nnz_idx, nnz_diff = _active_vars(
                    x, g, nnz_old, opt_tol, weights, dual_norm
                )
                if not nnz_diff:
                    if nnz_x == nnz_g:
                        iter_lim_lsqr = 20
                    else:
                        iter_lim_lsqr = 5
                    nnz_idx = np.abs(x) >= opt_tol

                    ebar = np.sign(x[nnz_idx])
                    nebar = np.size(ebar)
                    Sprod = _LSQRprod(A, nnz_idx, ebar, n)

                    dxbar, istop, niters_lsqr = lsqr(
                        Sprod, r, 1e-5, 1e-1, 1e-1, 1e12, iter_lim=iter_lim_lsqr, show=0
                    )[0:3]
                    nprodA += niters_lsqr
                    nprodAt += niters_lsqr + 1
                    niters_lsqr = niters_lsqr + niters_lsqr

                    # LSQR iterations successful. Take the subspace step.
                    if istop != 4:
                        # Push dx back into full space: dx = Z dx.
                        dx = np.zeros(n, dtype=x.dtype)
                        dx[nnz_idx] = dxbar - (1 / nebar) * np.dot(
                            np.dot(np.conj(ebar.T), dxbar), dxbar
                        )

                        # Find largest step to a change in sign.
                        block1 = nnz_idx & (x < 0) & (dx > +piv_tol)
                        block2 = nnz_idx & (x > 0) & (dx < -piv_tol)
                        alpha1 = np.inf
                        alpha2 = np.inf
                        if np.any(block1):
                            alpha1 = min(-x[block1] / dx[block1])
                        if np.any(block2):
                            alpha2 = min(-x[block2] / dx[block2])
                        alpha = min([1, alpha1, alpha2])
                        if alpha < 0:
                            raise ValueError("Alpha smaller than zero")
                        if np.dot(np.conj(ebar.T), dx[nnz_idx]) > opt_tol:
                            raise ValueError(
                                "Subspace update signed sum " "bigger than tolerance"
                            )
                        # Update variables.
                        x = x + alpha * dx
                        start_time_matvec = time.time()
                        r = b - A.matvec(x)
                        time_matprod += time.time() - start_time_matvec
                        f = abs(np.dot(np.conj(r), r)) / 2.0
                        subspace = True
                        nprodA += 1

                if primal_norm(x, weights) > tau + opt_tol:
                    raise ValueError("Primal norm out of bound")

            # Update gradient and compute new Barzilai-Borwein scaling.
            if not lnerr:
                start_time_matvec = time.time()
                g = -A.rmatvec(r)
                time_matprod += time.time() - start_time_matvec
                nprodAt += 1
                s = x - xold
                y = g - gold
                sts = np.dot(np.conj(s), s)
                sty = np.dot(np.conj(s), y)
                if sty <= 0:
                    gstep = step_max
                else:
                    gstep = min(step_max, max(step_min, sts / sty))
            else:
                gstep = min(step_max, gstep)
            break  # Leave while loop. This is done to allow stopping the
            # computations at any time within the loop if max_matvec is
            # reached. If this is not the case, the loop is stopped here.

        if stat == EXIT_MATVEC_LIMIT:
            niters -= 1
            x = xold.copy()
            f = fold
            g = gold.copy()
            r = rold.copy()
            break

        #  Update function history.
        if single_tau or f > sigma**2 / 2.0:  # Dont update if superoptimal.
            last_fv[np.mod(niters, n_prev_vals)] = f.copy()
            if fbest > f:
                fbest = f.copy()
                xbest = x.copy()

    # Restore best solution (only if solving single problem).
    if single_tau and f > fbest:
        rnorm = np.sqrt(2.0 * fbest)
        print("Restoring best iterate to objective " + str(rnorm))
        x = xbest.copy()
        start_time_matvec = time.time()
        r = b - A.matvec(x)
        g = -A.rmatvec(r)
        time_matprod += time.time() - start_time_matvec
        gnorm = dual_norm(g, weights)
        rnorm = np.linalg.norm(r)
        nprodA += 1
        nprodAt += 1

    # Final cleanup before exit.
    info = {}
    info["tau"] = tau
    info["rnorm"] = rnorm
    info["rgap"] = rgap
    info["gnorm"] = gnorm
    info["stat"] = stat
    info["niters"] = niters
    info["nprodA"] = nprodA
    info["nprodAt"] = nprodAt
    info["n_newton"] = n_newton
    info["time_project"] = time_project
    info["time_matprod"] = time_matprod
    info["niters_lsqr"] = niters_lsqr
    info["time_total"] = time.time() - start_time
    info["xnorm1"] = xnorm1[0:niters]
    info["rnorm2"] = rnorm2[0:niters]
    info["lambdaa"] = lambdaa[0:niters]

    # Print final output.
    if verbosity >= 1:
        _printf(fid, "")
        if stat == EXIT_OPTIMAL:
            _printf(fid, "EXIT -- Optimal solution found")
        elif stat == EXIT_ITERATIONS:
            _printf(fid, "ERROR EXIT -- Too many iterations")
        elif stat == EXIT_ROOT_FOUND:
            _printf(fid, "EXIT -- Found a root")
        elif stat == EXIT_BPSOL_FOUND:
            _printf(fid, "EXIT -- Found a BP solution")
        elif stat == EXIT_LEAST_SQUARES:
            _printf(fid, "EXIT -- Found a least-squares solution")
        elif stat == EXIT_LINE_ERROR:
            _printf(fid, "ERROR EXIT -- Linesearch error (%d)" % lnerr)
        elif stat == EXIT_SUBOPTIMAL_BP:
            _printf(fid, "EXIT -- Found a suboptimal BP solution")
        elif stat == EXIT_MATVEC_LIMIT:
            _printf(fid, "EXIT -- Maximum matrix-vector operations reached")
        elif stat == EXIT_ACTIVE_SET:
            _printf(fid, "EXIT -- Found a possible active set")
        else:
            _printf(fid, "SPGL1 ERROR: Unknown termination condition")
        _printf(fid, "")

        _printf(
            fid,
            "%-20s:  %6i %6s %-20s:  %6.1f"
            % (
                "Products with A",
                nprodA,
                "",
                "Total time   (secs)",
                info["time_total"],
            ),
        )
        _printf(
            fid,
            "%-20s:  %6i %6s %-20s:  %6.1f"
            % (
                "Products with A^H",
                nprodAt,
                "",
                "Project time (secs)",
                info["time_project"],
            ),
        )
        _printf(
            fid,
            "%-20s:  %6i %6s %-20s:  %6.1f"
            % (
                "Newton iterations",
                n_newton,
                "",
                "Mat-vec time (secs)",
                info["time_matprod"],
            ),
        )
        _printf(
            fid,
            "%-20s:  %6i %6s %-20s:  %6i"
            % ("Line search its", nline_tot, "", "Subspace iterations", niters_lsqr),
        )

    return x, r, g, info


def spg_bp(A, b, **kwargs):
    """Basis pursuit (BP) problem.

    ``spg_bp`` is designed to solve the basis pursuit problem::

        (BP)  minimize  ||x||_1  subject to  Ax = b,

    where ``A`` is an M-by-N matrix, ``b`` is an M-vector.
    ``A`` can be an explicit M-by-N matrix or a
    :class:`scipy.sparse.linalg.LinearOperator`.

    This is equivalent to calling ``spgl1(A, b, tau=0, sigma=0)

    Parameters
    ----------
    A : {sparse matrix, ndarray, LinearOperator}
        Representation of an m-by-n matrix.  It is required that
        the linear operator can produce ``Ax`` and ``A^T x``.
    b : array_like, shape (m,)
        Right-hand side vector ``b``.
    kwargs : dict, optional
        Additional input parameters (refer to :func:`spgl1.spgl1` for a list
        of possible parameters)

    Returns
    -------
    x : array_like, shape (n,)
        Inverted model
    r : array_like, shape (m,)
        Final residual
    g : array_like, shape (h,)
        Final gradient
    info : dict
        See splg1.
    """
    sigma = 0
    tau = 0
    x0 = None
    x, r, g, info = spgl1(A, b, tau, sigma, x0, **kwargs)
    return x, r, g, info


def spg_bpdn(A, b, sigma, **kwargs):
    """Basis pursuit denoise (BPDN) problem.


    ``spg_bpdn`` is designed to solve the basis pursuit denoise problem::

        (BPDN)  minimize  ||x||_1  subject to  ||A x - b|| <= sigma

    where ``A`` is an M-by-N matrix, ``b`` is an M-vector.
    ``A`` can be an explicit M-by-N matrix or a
    :class:`scipy.sparse.linalg.LinearOperator`.

    This is equivalent to calling ``spgl1(A, b, tau=0, sigma=sigma)

    Parameters
    ----------
    A : {sparse matrix, ndarray, LinearOperator}
        Representation of an m-by-n matrix.  It is required that
        the linear operator can produce ``Ax`` and ``A^T x``.
    b : array_like, shape (m,)
        Right-hand side vector ``b``.
    kwargs : dict, optional
        Additional input parameters (refer to :func:`spgl1.spgl1` for a list
        of possible parameters)

    Returns
    -------
    x : array_like, shape (n,)
        Inverted model
    r : array_like, shape (m,)
        Final residual
    g : array_like, shape (h,)
        Final gradient
    info : dict
        See spgl1.

    """
    tau = 0
    x0 = None
    x, r, g, info = spgl1(A, b, tau, sigma, x0, **kwargs)
    return x, r, g, info


def spg_lasso(A, b, tau, **kwargs):
    """LASSO problem.

    ``spg_lasso`` is designed to solve the Lasso problem::

        (LASSO)  minimize  ||Ax - b||_2  subject to  ||x||_1 <= tau

    where ``A`` is an M-by-N matrix, ``b`` is an M-vector.
    ``A`` can be an explicit M-by-N matrix or a
    :class:`scipy.sparse.linalg.LinearOperator`.

    This is equivalent to calling ``spgl1(A, b, tau=tau, sigma=0)

    Parameters
    ----------
    A : {sparse matrix, ndarray, LinearOperator}
        Representation of an m-by-n matrix.  It is required that
        the linear operator can produce ``Ax`` and ``A^T x``.
    b : array_like, shape (m,)
        Right-hand side vector ``b``.
    kwargs : dict, optional
        Additional input parameters (refer to :func:`spgl1.spgl1` for a list
        of possible parameters)

    Returns
    -------
    x : array_like, shape (n,)
        Inverted model
    r : array_like, shape (m,)
        Final residual
    g : array_like, shape (h,)
        Final gradient
    info : dict
        See spgl1.

    """
    sigma = 0
    x0 = None
    x, r, g, info = spgl1(A, b, tau, sigma, x0, **kwargs)
    return x, r, g, info


def spg_mmv(A, B, sigma=0, **kwargs):
    """MMV problem.

    ``spg_mmv`` is designed to solve the  multi-measurement vector
    basis pursuit denoise::

        (MMV)  minimize  ||X||_1,2  subject to  ||A X - B||_2,2 <= sigma

    where ``A`` is an M-by-N matrix, ``b`` is an M-by-G matrix, and ```sigma``
    is a nonnegative scalar. ``A`` can be an explicit M-by-N matrix or a
    :class:`scipy.sparse.linalg.LinearOperator`.

    Parameters
    ----------
    A : {sparse matrix, ndarray, LinearOperator}
        Representation of an M-by-N  matrix.  It is required that
        the linear operator can produce ``Ax`` and ``A^T x``.
    b : array_like, shape (m,)
        Right-hand side matrix ``b`` of size M-by-G.
    sigma : float, optional
        BPDN threshold. If different from ``None``, spgl1 solves BPDN problem
    kwargs : dict, optional
        Additional input parameters (refer to :func:`spgl1.spgl1` for a list
        of possible parameters)

    Returns
    -------
    x : array_like, shape (n,)
        Inverted model
    r : array_like, shape (m,)
        Final residual
    g : array_like, shape (h,)
        Final gradient
    info : dict
        See spgl1.

    """
    A = aslinearoperator(A)
    m, n = A.shape
    groups = B.shape[1]
    A = _blockdiag(A, m, n, groups)

    # Set projection specific functions
    _primal_norm = (
        _norm_l12_primal
        if "primal_norm" not in kwargs.keys()
        else kwargs["primal_norm"]
    )
    _dual_norm = (
        _norm_l12_dual if "dual_norm" not in kwargs.keys() else kwargs["dual_norm"]
    )
    _project = (
        _norm_l12_project if "project" not in kwargs.keys() else kwargs["project"]
    )
    kwargs.pop("primal_norm", None)
    kwargs.pop("dual_norm", None)
    kwargs.pop("project", None)

    project = lambda x, weight, tau: _project(groups, x, weight, tau)
    primal_norm = lambda x, weight: _primal_norm(groups, x, weight)
    dual_norm = lambda x, weight: _dual_norm(groups, x, weight)

    tau = 0
    x0 = None
    x, r, g, info = spgl1(
        A,
        B.ravel(),
        tau,
        sigma,
        x0,
        project=project,
        primal_norm=primal_norm,
        dual_norm=dual_norm,
        **kwargs
    )
    x = x.reshape(n, groups)
    g = g.reshape(n, groups)

    return x, r, g, info



import numpy as np
import pytest
from numpy.testing import assert_array_almost_equal
from scipy.sparse import csr_matrix, spdiags
from scipy.sparse.linalg import lsqr as splsqr

from spgl1 import spg_bp, spg_bpdn, spg_lasso, spg_mmv
from spgl1.lsqr import lsqr
from spgl1.spgl1 import (
    norm_l12nn_dual,
    norm_l12nn_primal,
    norm_l12nn_project,
    oneprojector,
)

# dense matrix (dtype=np.float64)
par1 = {
    "n": 50,
    "m": 50,
    "k": 14,
    "dtype": np.float64,
    "sparse": False,
}  # square
par2 = {
    "n": 128,
    "m": 50,
    "k": 14,
    "dtype": np.float64,
    "sparse": False,
}  # underdetermined
par3 = {
    "n": 50,
    "m": 100,
    "k": 14,
    "dtype": np.float64,
    "sparse": False,
}  # overdetermined

# sparse matrix (dtype=np.float64)
par1_sp = {
    "n": 50,
    "m": 50,
    "k": 14,
    "dtype": np.float64,
    "sparse": True,
}  # square
par2_sp = {
    "n": 128,
    "m": 50,
    "k": 14,
    "dtype": np.float64,
    "sparse": True,
}  # underdetermined
par3_sp = {
    "n": 50,
    "m": 100,
    "k": 14,
    "dtype": np.float64,
    "sparse": True,
}  # overdetermined

# dense matrix (dtype=np.float32)
par1_f32 = {
    "n": 50,
    "m": 50,
    "k": 14,
    "dtype": np.float32,
    "sparse": False,
}  # square
par2_f32 = {
    "n": 128,
    "m": 50,
    "k": 14,
    "dtype": np.float32,
    "sparse": False,
}  # underdetermined
par3_f32 = {
    "n": 50,
    "m": 100,
    "k": 14,
    "dtype": np.float32,
    "sparse": False,
}  # overdetermined


@pytest.mark.parametrize("par", [(par1), (par1_f32)])
def test_oneprojector(par):
    """One projector with tau=1

    minimize_x  ||b-x||_2  subject to  ||x||_1 <= 1.
    """
    np.random.seed(0)

    b = np.random.normal(0.0, 1.0, par["n"]).astype(dtype=par["dtype"])
    d = 1.0
    bp = oneprojector(b, d, 1.0)
    assert bp.dtype == par["dtype"]
    assert np.linalg.norm(bp, 1) < 1.0


@pytest.mark.parametrize(
    "par",
    [
        (par1),
        (par2),
        (par3),
        (par1_sp),
        (par2_sp),
        (par3_sp),
        (par1_f32),
        (par2_f32),
        (par3_f32),
    ],
)
def test_lasso(par):
    """LASSO problem for ||x||_1 <= pi:

    minimize ||Ax-b||_2 subject to ||x||_1 <= 3.14159...

    """
    np.random.seed(0)

    n = par["n"]
    m = par["m"]
    k = par["k"]

    A, A1 = np.linalg.qr(np.random.randn(n, m), "reduced")
    if m > n:
        A = A1.copy()
    if par["sparse"]:
        A = csr_matrix(A)
    A = A.astype(par["dtype"])

    # Create sparse vector
    p = np.random.permutation(m)
    p = p[0:k]
    x = np.zeros(m, dtype=par["dtype"])
    x[p] = np.random.randn(k)

    # Set up vector b, and run solver
    b = A.dot(x)
    tau = np.pi

    xinv, resid, _, _ = spg_lasso(A, b, tau, verbosity=0)
    assert xinv.dtype == par["dtype"]
    assert np.linalg.norm(xinv, 1) - np.pi < 1e-5

    # Run solver with subspace minimization
    xinv, resid, _, _ = spg_lasso(A, b, tau, subspace_min=True, verbosity=0)
    assert xinv.dtype == par["dtype"]
    assert np.linalg.norm(xinv, 1) - np.pi < 1e-5


@pytest.mark.parametrize(
    "par",
    [
        (par1),
        (par2),
        (par3),
        (par1_sp),
        (par2_sp),
        (par3_sp),
        (par1_f32),
        (par2_f32),
        (par3_f32),
    ],
)
def test_bp(par):
    """Basis pursuit (BP) problem:

    minimize ||x||_1 subject to Ax = b

    """
    np.random.seed(0)

    # Create random m-by-n encoding matrix
    n = par["n"]
    m = par["m"]
    k = par["k"]

    A, A1 = np.linalg.qr(np.random.randn(n, m), "reduced")
    if m > n:
        A = A1.copy()
    A = A.astype(par["dtype"])

    # Create sparse vector
    p = np.random.permutation(m)
    p = p[0:k]
    x = np.zeros(m, dtype=par["dtype"])
    x[p] = np.random.randn(k)

    # Set up vector b, and run solver
    b = A.dot(x)
    xinv, _, _, _ = spg_bp(A, b, verbosity=0)
    assert xinv.dtype == par["dtype"]
    assert_array_almost_equal(x, xinv, decimal=3 if par["dtype"] == np.float64 else 2)

    # Run solver with subspace minimization
    xinv, _, _, _ = spg_bp(A, b, subspace_min=True, verbosity=0)
    assert xinv.dtype == par["dtype"]
    assert_array_almost_equal(x, xinv, decimal=3 if par["dtype"] == np.float64 else 2)


@pytest.mark.parametrize(
    "par", [(par1), (par3), (par1_sp), (par3_sp), (par1_f32), (par3_f32)]
)
def test_bpdn(par):
    """Basis pursuit denoise (BPDN) problem:

    minimize ||x||_1 subject to ||Ax - b||_2 <= 0.1

    """
    np.random.seed(0)

    # Create random m-by-n encoding matrix
    n = par["n"]
    m = par["m"]
    k = par["k"]

    A, A1 = np.linalg.qr(np.random.randn(n, m), "reduced")
    if m > n:
        A = A1.copy()
    A = A.astype(par["dtype"])

    # Create sparse vector
    p = np.random.permutation(m)
    p = p[0:k]
    x = np.zeros(m, dtype=par["dtype"])
    x[p] = np.random.randn(k)

    # Set up vector b, and run solver
    b = A.dot(x) + np.random.randn(n).astype(par["dtype"]) * 0.075

    sigma = 0.10
    xinv, resid, _, _ = spg_bpdn(A, b, sigma, iter_lim=1000, verbosity=0)
    assert xinv.dtype == par["dtype"]
    assert (
        np.linalg.norm(resid) < sigma * 1.1
    )  # need to check why resid is slighly bigger than sigma

    # Run solver with subspace minimization
    xinv, _, _, _ = spg_bpdn(A, b, sigma, subspace_min=True, verbosity=0)
    assert xinv.dtype == par["dtype"]
    assert (
        np.linalg.norm(resid) < sigma * 1.1
    )  # need to check why resid is slighly bigger than sigma


@pytest.mark.parametrize(
    "par",
    [
        (par1),
        (par2),
        (par3),
        (par1_sp),
        (par2_sp),
        (par3_sp),
        (par1_f32),
        (par2_f32),
        (par3_f32),
    ],
)
def test_bp_complex(par):
    """Basis pursuit (BP) problem for complex variables:

    minimize ||x||_1 subject to Ax = b

    """
    np.random.seed(0)

    dtypecomplex = (
        np.ones(1, dtype=par["dtype"]) + 1j * np.ones(1, dtype=par["dtype"])
    ).dtype

    # Create random m-by-n encoding matrix
    n = par["n"]
    m = par["m"]
    k = par["k"]

    A, A1 = np.linalg.qr(np.random.randn(n, m), "reduced")
    if m > n:
        A = A1.copy()
    A = A.astype(dtypecomplex)

    # Create sparse vector
    p = np.random.permutation(m)
    p = p[0:k]
    x = np.zeros(m, dtype=dtypecomplex)
    x[p] = np.random.randn(k) + 1j * np.random.randn(k)

    # Set up vector b, and run solver
    b = A.dot(x)
    xinv, _, _, _ = spg_bp(A, b, verbosity=0)
    assert xinv.dtype == dtypecomplex
    assert_array_almost_equal(x, xinv, decimal=3 if par["dtype"] == np.float64 else 2)

    # Run solver with subspace minimization
    xinv, _, _, _ = spg_bp(A, b, subspace_min=True, verbosity=0)
    assert xinv.dtype == dtypecomplex
    assert_array_almost_equal(x, xinv, decimal=3 if par["dtype"] == np.float64 else 2)


@pytest.mark.parametrize(
    "par",
    [
        (par1),
        (par2),
        (par3),
        (par1_sp),
        (par2_sp),
        (par3_sp),
        (par1_f32),
        (par2_f32),
        (par3_f32),
    ],
)
def test_weighted_bp(par):
    """Weighted Basis pursuit (WBP) problem:

    minimize ||y||_1 subject to AW^{-1}y = b

    or

    minimize ||Wx||_1 subject to Ax = b

    """
    np.random.seed(0)

    # Create random m-by-n encoding matrix
    n = par["n"]
    m = par["m"]
    k = par["k"]

    A, A1 = np.linalg.qr(np.random.randn(n, m), "reduced")
    if m > n:
        A = A1.copy()
    A = A.astype(par["dtype"])

    # Create sparse vector
    p = np.random.permutation(m)
    p = p[0:k]
    x = np.zeros(m, dtype=par["dtype"])
    x[p] = np.random.randn(k)

    # Set up weights w and vector b
    w = 0.1 * np.random.rand(m).astype(dtype=par["dtype"]) + 0.1  # Weights
    b = A.dot(x / w)  # Signal
    xinv, _, _, _ = spg_bp(A, b, iter_lim=1000, weights=w, verbosity=0)

    # Reconstructed solution, with weighting
    xinv *= w
    assert xinv.dtype == par["dtype"]
    assert_array_almost_equal(x, xinv, decimal=3 if par["dtype"] == np.float64 else 2)


@pytest.mark.parametrize(
    "par",
    [
        (par1),
        (par2),
        (par3),
        (par1_sp),
        (par2_sp),
        (par3_sp),
        (par1_f32),
        (par2_f32),
        (par3_f32),
    ],
)
def test_multiple_measurements(par):
    """Multiple measurement vector (MMV) problem:

    minimize ||Y||_1,2 subject to AW^{-1}Y = B

    """
    np.random.seed(0)

    # Create random m-by-n encoding matrix
    m = par["m"]
    n = par["n"]
    k = par["k"]
    l = 6

    A = np.random.randn(m, n)
    A = A.astype(par["dtype"])

    p = np.random.permutation(n)[:k]
    X = np.zeros((n, l), dtype=par["dtype"])
    X[p, :] = np.random.randn(k, l)

    weights = 0.1 * np.random.rand(n).astype(par["dtype"]) + 0.1
    W = 1 / weights * np.eye(n, dtype=par["dtype"])

    B = A.dot(W).dot(X)

    # Solve unweighted version
    X_uw, _, _, _ = spg_mmv(A.dot(W), B, 0, verbosity=0)

    # Solve weighted version
    X_w, _, _, _ = spg_mmv(A, B, 0, weights=weights, verbosity=0)
    X_w = spdiags(weights, 0, n, n).dot(X_w)

    assert X_uw.dtype == par["dtype"]
    assert X_w.dtype == par["dtype"]
    assert_array_almost_equal(X, X_uw, decimal=1)
    assert_array_almost_equal(X, X_w, decimal=1)


@pytest.mark.parametrize(
    "par",
    [
        (par1),
        (par2),
        (par3),
        (par1_sp),
        (par2_sp),
        (par3_sp),
        (par1_f32),
        (par2_f32),
        (par3_f32),
    ],
)
def test_multiple_measurements_nonnegative(par):
    """Multiple measurement vector (MMV) problem with non-negative norm:"""
    np.random.seed(0)

    # Create random m-by-n encoding matrix
    m = par["m"]
    n = par["n"]
    k = par["k"]
    l = 6

    A = np.random.randn(m, n)
    A = A.astype(par["dtype"])

    p = np.random.permutation(n)[:k]
    X = np.zeros((n, l), dtype=par["dtype"])
    X[p, :] = np.abs(np.random.randn(k, l))

    B = A.dot(X)

    Xnn, _, _, _ = spg_mmv(
        A,
        B,
        0,
        project=norm_l12nn_project,
        primal_norm=norm_l12nn_primal,
        dual_norm=norm_l12nn_dual,
        iter_lim=20,
        verbosity=0,
    )
    assert Xnn.dtype == par["dtype"]
    assert np.any(Xnn < 0) == False


@pytest.mark.parametrize("par", [(par1), (par3), (par1_sp), (par3_sp)])
def test_lsqr(par):
    """Compare local LSQR with scipy LSQR"""

    def Aprodfun(A, x, mode):
        if mode == 1:
            y = np.dot(A, x)
        else:
            return np.dot(np.conj(A.T), x)
        return y

    np.random.seed(0)

    # Create random m-by-n encoding matrix
    m = par["m"]
    n = par["n"]
    A = np.random.normal(0, 1, (m, n))
    Aprod = lambda x, mode: Aprodfun(A, x, mode)
    x = np.ones(n)
    y = A.dot(x)

    damp = 1e-10
    atol = 1e-10
    btol = 1e-10
    conlim = 1e12
    itn_max = 500
    show = 0

    xinv, istop, itn, r1norm, r2norm, anorm, acond, arnorm, xnorm, var = lsqr(
        m, n, Aprod, y, damp, atol, btol, conlim, itn_max, show
    )

    (
        xinv_sp,
        istop_sp,
        itn_sp,
        r1norm_sp,
        r2norm_sp,
        anorm_sp,
        acond_sp,
        arnorm_sp,
        xnorm_sp,
        var_sp,
    ) = splsqr(A, y, damp, atol, btol, conlim, itn_max, show)

    assert_array_almost_equal(xinv, x, decimal=3 if par["dtype"] == np.float64 else 2)
    assert_array_almost_equal(
        xinv_sp, x, decimal=3 if par["dtype"] == np.float64 else 2
    )



r"""
SPGL1 Tutorial
==============
In this tutorial we will explore the different solvers in the ``spgl1``
package and apply them to different toy examples.
"""
import matplotlib.pyplot as plt
import numpy as np
from scipy.sparse import spdiags
from scipy.sparse.linalg import LinearOperator

import spgl1

# Initialize random number generators
np.random.seed(43273289)

###############################################################################
# Create random m-by-n encoding matrix and sparse vector
m = 50
n = 128
k = 14
[A, Rtmp] = np.linalg.qr(np.random.randn(n, m), "reduced")
A = A.T
p = np.random.permutation(n)
p = p[0:k]
x0 = np.zeros(n)
x0[p] = np.random.randn(k)

###############################################################################
# Solve the underdetermined LASSO problem for :math:`||\mathbf{x}||_1 <= \pi`:
#
# .. math::
#     min.||\mathbf{Ax}-\mathbf{b}||_2 \quad subject \quad  to \quad
#     ||\mathbf{x}||_1 <= \pi
b = A.dot(x0)
tau = np.pi
x, resid, grad, info = spgl1.spg_lasso(A, b, tau, verbosity=1)

print("%s%s%s" % ("-" * 35, " Solution ", "-" * 35))
print(
    "nonzeros(x) = %i,   ||x||_1 = %12.6e,   ||x||_1 - pi = %13.6e"
    % (np.sum(np.abs(x) > 1e-5), np.linalg.norm(x, 1), np.linalg.norm(x, 1) - np.pi)
)
print("%s" % ("-" * 80))

###############################################################################
# Solve the basis pursuit (BP) problem:
#
#  .. math::
#     min.  ||\mathbf{x}||_1 \quad subject \quad  to \quad
#     \mathbf{Ax} = \mathbf{b}
b = A.dot(x0)
x, resid, grad, info = spgl1.spg_bp(A, b, verbosity=2)

plt.figure()
plt.plot(x, "b")
plt.plot(x0, "ro")
plt.legend(("Recovered coefficients", "Original coefficients"))
plt.title("Basis Pursuit")

plt.figure()
plt.plot(info["xnorm1"], info["rnorm2"], ".-k")
plt.xlabel(r"$||x||_1$")
plt.ylabel(r"$||r||_2$")
plt.title("Pareto curve")

plt.figure()
plt.plot(
    np.arange(info["niters"]),
    info["rnorm2"] / max(info["rnorm2"]),
    ".-k",
    label=r"$||r||_2$",
)
plt.plot(
    np.arange(info["niters"]),
    info["xnorm1"] / max(info["xnorm1"]),
    ".-r",
    label=r"$||x||_1$",
)
plt.xlabel(r"#iter")
plt.ylabel(r"$||r||_2/||x||_1$")
plt.title("Norms")
plt.legend()

###############################################################################
# Solve the basis pursuit denoise (BPDN) problem:
#
# .. math::
#     min. ||\mathbf{x}||_1 \quad subject \quad to \quad ||\mathbf{Ax} -
#     \mathbf{b}||_2 <= 0.1

b = A.dot(x0) + np.random.randn(m) * 0.075
sigma = 0.10  # Desired ||Ax - b||_2
x, resid, grad, info = spgl1.spg_bpdn(A, b, sigma, iter_lim=100, verbosity=2)

plt.figure()
plt.plot(x, "b")
plt.plot(x0, "ro")
plt.legend(("Recovered coefficients", "Original coefficients"))
plt.title("Basis Pursuit Denoise")

###############################################################################
# Solve the basis pursuit (BP) problem in complex variables:
#
#  .. math::
#      min. ||\mathbf{z}||_1 \quad subject \quad to \quad
#      \mathbf{Az} = \mathbf{b}
class partialFourier(LinearOperator):
    def __init__(self, idx, n):
        self.idx = idx
        self.n = n
        self.shape = (len(idx), n)
        self.dtype = np.complex128

    def _matvec(self, x):
        # % y = P(idx) * FFT(x)
        z = np.fft.fft(x) / np.sqrt(n)
        return z[idx]

    def _rmatvec(self, x):
        z = np.zeros(n, dtype=complex)
        z[idx] = x
        return np.fft.ifft(z) * np.sqrt(n)


# Create partial Fourier operator with rows idx
idx = np.random.permutation(n)
idx = idx[0:m]
opA = partialFourier(idx, n)

# Create sparse coefficients and b = A * z0
z0 = np.zeros(n, dtype=complex)
z0[p] = np.random.randn(k) + 1j * np.random.randn(k)
b = opA.matvec(z0)

# Solve problem
z, resid, grad, info = spgl1.spg_bp(opA, b, verbosity=2)

plt.figure()
plt.plot(z.real, "b+", markersize=15.0)
plt.plot(z0.real, "bo")
plt.plot(z.imag, "r+", markersize=15.0)
plt.plot(z0.imag, "ro")
plt.legend(
    ("Recovered (real)", "Original (real)", "Recovered (imag)", "Original (imag)")
)
plt.title("Complex Basis Pursuit")

###############################################################################
# We can also sample the Pareto frontier at 100 points:
#
#  .. math::
#      \phi(\tau) = min. ||\mathbf{Ax}-\mathbf{b}||_2 \quad subject \quad
#      to \quad ||\mathbf{x}|| <= \tau
b = A.dot(x0)
x = np.zeros(n)
tau = np.linspace(0, 1.05 * np.linalg.norm(x0, 1), 100)
tau[0] = 1e-10
phi = np.zeros(tau.size)

for i in range(tau.size):
    x, r, grad, info = spgl1.spgl1(A, b, tau[i], 0, x, iter_lim=1000)
    phi[i] = np.linalg.norm(r)

plt.figure()
plt.plot(tau, phi, ".")
plt.title("Pareto frontier")
plt.xlabel("||x||_1")
plt.ylabel("||Ax-b||_2")

###############################################################################
# We now solve the weighted basis pursuit (BP) problem:
#
#  .. math::
#     min. ||\mathbf{y}||_1 \quad subject \quad  to \quad \mathbf{AW}^{-1}\mathbf{y} = \mathbf{b}
#
# and
#
#  .. math::
#     min. ||\mathbf{Wx}||_1 \quad subject \quad to \quad \mathbf{Ax} = \mathbf{b}
#
# followed by setting :math`\mathbf{y} = \mathbf{Wx}`.

# Sparsify vector x0 a bit more to get exact recovery
k = 9
x0 = np.zeros(n)
x0[p[0:k]] = np.random.randn(k)

# Set up weights w and vector b
w = np.random.rand(n) + 0.1  # Weights
b = A.dot(x0 / w)  # Signal

# Solve problem
x, resid, grad, info = spgl1.spg_bp(A, b, iter_lim=1000, weights=w)

# Reconstructed solution, with weighting
x1 = x * w

plt.figure()
plt.plot(x1, "b")
plt.plot(x0, "ro")
plt.legend(("Coefficients", "Original coefficients"))
plt.title("Weighted Basis Pursuit")
k = 9
x0 = np.zeros(n)
x0[p[0:k]] = np.random.randn(k)

###############################################################################
# Finally we solve the multiple measurement vector (MMV) problem
#
#  .. math::
#     min. | | \mathbf{Y} | |_{1, 2}  \quad subject \quad to \quad
#     \mathbf{AW}^{-1} \mathbf{Y} = \mathbf{B}
#
# and the weighted MMV problem(weights on the rows of X):
#
#  .. math::
#     min. | | \mathbf{WX} | |_{1, 2} \quad subject \quad
#     to \quad \mathbf{AX} = \mathbf{B}
#
# followed by setting :math:`\mathbf{Y} = \mathbf{WX}`.

# Create problem
m = 100
n = 150
k = 12
l = 6
A = np.random.randn(m, n)
p = np.random.permutation(n)[:k]
X0 = np.zeros((n, l))
X0[p, :] = np.random.randn(k, l)

weights = 3 * np.random.rand(n) + 0.1
W = 1 / weights * np.eye(n)
B = A.dot(W).dot(X0)

# Solve unweighted version
x_uw, _, _, _ = spgl1.spg_mmv(A.dot(W), B, 0, verbosity=1)

# Solve weighted version
x_w, _, _, _ = spgl1.spg_mmv(A, B, 0, weights=weights, verbosity=2)
x_w = spdiags(weights, 0, n, n).dot(x_w)

# Plot results
plt.figure()
plt.plot(x_uw[:, 0], "b-", label="Coefficients (1)")
plt.plot(x_w[:, 0], "b.", label="Coefficients (2)")
plt.plot(X0[:, 0], "ro", label="Original coefficients")
plt.legend()
plt.title("Weighted Basis Pursuit with Multiple Measurement Vectors")

plt.figure()
plt.plot(x_uw[:, 1], "b-", label="Coefficients (1)")
plt.plot(x_w[:, 1], "b.", label="Coefficients (2)")
plt.plot(X0[:, 1], "ro", label="Original coefficients")
plt.legend()
plt.title("Weighted Basis Pursuit with Multiple Measurement Vectors")



# -*- coding: utf-8 -*-
import sys
import os
import datetime
import sphinx_rtd_theme
from sphinx_gallery.sorting import ExampleTitleSortKey
from pkg_resources import get_distribution

# Sphinx needs to be able to import the package to use autodoc and get the version number
sys.path.insert(0, os.path.abspath('../../spgl1'))

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.autosummary',
    'sphinx.ext.coverage',
    'sphinx.ext.mathjax',
    'sphinx.ext.doctest',
    'sphinx.ext.viewcode',
    'sphinx.ext.extlinks',
    "sphinx.ext.intersphinx",
    'matplotlib.sphinxext.plot_directive',
    'numpydoc',
    'nbsphinx',
    'sphinx_gallery.gen_gallery',
    #'sphinx.ext.napoleon',
]

# intersphinx configuration
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "numpy": ("https://docs.scipy.org/doc/numpy/", None),
    "scipy": ("https://docs.scipy.org/doc/scipy/reference", None),
    "sklearn": ("http://scikit-learn.org/stable/", None),
    "pandas": ("http://pandas.pydata.org/pandas-docs/stable/", None),
    "matplotlib": ("https://matplotlib.org/", None),
}

## Generate autodoc stubs with summaries from code
autosummary_generate = True

## Include Python objects as they appear in source files
autodoc_member_order = 'bysource'

## Default flags used by autodoc directives
autodoc_default_flags = ['members']

numpydoc_show_class_members = False
numpydoc_show_inherited_class_members = False
numpydoc_class_members_toctree = False

sphinx_gallery_conf = {
    # path to your examples scripts
    'examples_dirs': ['../../tutorials',],
    # path where to save gallery generated examples
    'gallery_dirs': ['tutorials'],
    'filename_pattern': '\.py',
    # Remove the "Download all examples" button from the top level gallery
    'download_all_examples': False,
    # Sort gallery example by file name instead of number of lines (default)
    'within_subsection_order': ExampleTitleSortKey,
    # directory where function granular galleries are stored
    'backreferences_dir': 'api/generated/backreferences',
    # Modules for which function level galleries are created.
    'doc_module': 'spgl1',
    # Insert links to documentation of objects in the examples
    'reference_url': {'spgl1': None}
}

# Always show the source code that generates a plot
plot_include_source = True
plot_formats = ['png']

# Sphinx project configuration
templates_path = ['_templates']
exclude_patterns = ['_build', '**.ipynb_checkpoints']
source_suffix = '.rst'

# The encoding of source files.
source_encoding = 'utf-8-sig'
master_doc = 'index'

# General information about the project
year = datetime.date.today().year
project = 'SPGL1'
copyright = '{}'.format(year)

# Version
version = get_distribution('spgl1').version
if len(version.split('+')) > 1 or version == 'unknown':
    version = 'dev'

# These enable substitutions using |variable| in the rst files
rst_epilog = """
.. |year| replace:: {year}
""".format(year=year)

html_last_updated_fmt = '%b %d, %Y'
html_title = 'SPGL1'
html_short_title = 'SPGL1'
html_static_path = ['_static']
html_extra_path = []
pygments_style = 'default'
add_function_parentheses = False
html_show_sourcelink = False
html_show_sphinx = True
html_show_copyright = True

# Theme config
html_theme = "sphinx_rtd_theme"
html_theme_options = {
    'logo_only': True,
    'display_version': True,
}
html_context = {
    'menu_links_name': 'Repository',
    'menu_links': [
        ('<i class="fa fa-github fa-fw"></i> Source Code', 'https://github.com/drrelyea/spgl1'),
    ],
    # Custom variables to enable "Improve this page"" and "Download notebook"
    # links
    'doc_path': 'docs/source',
    'galleries': sphinx_gallery_conf['gallery_dirs'],
    'gallery_dir': sphinx_gallery_conf['gallery_dirs'],
    'github_project': 'drrelyea',
    'github_repo': 'spgl1',
    'github_version': 'master',
}

# Load the custom CSS files (needs sphinx >= 1.6 for this to work)
def setup(app):
    app.add_css_file("style.css")



from __future__ import division

import numpy as np


def lsqr(m, n, Aprod, b, damp, atol, btol, conlim, itnlim, show):
    #
    # LSQR solves  Ax = b  or  min ||b - Ax||_2  if damp = 0,
    # or   min || (b)  -  (  A   )x ||   otherwise.
    #          || (0)     (damp I)  ||2
    # A  is an m by n matrix defined or a function handle of aprod( mode,x ),
    # that performs the matrix-vector operations.
    # If mode = 1,   aprod  must return  y = Ax   without altering x.
    # If mode = 2,   aprod  must return  y = A'x  without altering x.

    # -----------------------------------------------------------------------
    # LSQR uses an iterative (conjugate-gradient-like) method.
    # For further information, see
    # 1. C. C. Paige and M. A. Saunders (1982a).
    #    LSQR: An algorithm for sparse linear equations and sparse least squares,
    #    ACM TOMS 8(1), 43-71.
    # 2. C. C. Paige and M. A. Saunders (1982b).
    #    Algorithm 583.  LSQR: Sparse linear equations and least squares problems,
    #    ACM TOMS 8(2), 195-209.
    # 3. M. A. Saunders (1995).  Solution of sparse rectangular systems using
    #    LSQR and CRAIG, BIT 35, 588-604.
    #
    # Input parameters:
    # atol, btol  are stopping tolerances.  If both are 1.0e-9 (say),
    #             the final residual norm should be accurate to about 9 digits.
    #             (The final x will usually have fewer correct digits,
    #             depending on cond(A) and the size of damp.)
    # conlim      is also a stopping tolerance.  lsqr terminates if an estimate
    #             of cond(A) exceeds conlim.  For compatible systems Ax = b,
    #             conlim could be as large as 1.0e+12 (say).  For least-squares
    #             problems, conlim should be less than 1.0e+8.
    #             Maximum precision can be obtained by setting
    #             atol = btol = conlim = zero, but the number of iterations
    #             may then be excessive.
    # itnlim      is an explicit limit on iterations (for safety).
    # show = 1    gives an iteration log,
    # show = 0    suppresses output.
    #
    # Output parameters:
    # x           is the final solution.
    # istop       gives the reason for termination.
    # istop       = 1 means x is an approximate solution to Ax = b.
    #             = 2 means x approximately solves the least-squares problem.
    # r1norm      = norm(r), where r = b - Ax.
    # r2norm      = sqrt( norm(r)^2  +  damp^2 * norm(x)^2 )
    #             = r1norm if damp = 0.
    # anorm       = estimate of Frobenius norm of Abar = [  A   ].
    #                                                    [damp*I]
    # acond       = estimate of cond(Abar).
    # arnorm      = estimate of norm(A'*r - damp^2*x).
    # xnorm       = norm(x).
    # var         (if present) estimates all diagonals of (A'A)^{-1} (if damp=0)
    #             or more generally (A'A + damp^2*I)^{-1}.
    #             This is well defined if A has full column rank or damp > 0.
    #             (Not sure what var means if rank(A) < n and damp = 0.)
    #
    #
    #        1990: Derived from Fortran 77 version of LSQR.
    # 22 May 1992: bbnorm was used incorrectly.  Replaced by anorm.
    # 26 Oct 1992: More input and output parameters added.
    # 01 Sep 1994: Matrix-vector routine is now a parameter 'aprodname'.
    #              Print log reformatted.
    # 14 Jun 1997: show  added to allow printing or not.
    # 30 Jun 1997: var   added as an optional output parameter.
    # 07 Aug 2002: Output parameter rnorm replaced by r1norm and r2norm.
    #              Michael Saunders, Systems Optimization Laboratory,
    #              Dept of MS&E, Stanford University.
    # 03 Jul 2007: Modified 'aprodname' to A, which can either be an m by n
    #              matrix, or a function handle.
    #              Ewout van den Berg, University of British Columbia
    # 03 Jul 2007: Modified 'test2' condition, omitted 'test1'.
    #              Ewout van den Berg, University of British Columbia
    # -----------------------------------------------------------------------

    nprodA = 0
    nprodAT = 0

    # Initialize.
    msg = (
        "The exact solution is  x = 0                              ",
        "Ax - b is small enough, given atol, btol                  ",
        "The least-squares solution is good enough, given atol     ",
        "The estimate of cond(Abar) has exceeded conlim            ",
        "Ax - b is small enough for this machine                   ",
        "The least-squares solution is good enough for this machine",
        "Cond(Abar) seems to be too large for this machine         ",
        "The iteration limit has been reached                      ",
    )

    wantvar = True
    var = np.zeros(n)

    if show:
        print(" ")
        print("LSQR            Least-squares solution of  Ax = b")
        str1 = "The matrix A has %8g rows  and %8g cols" % (m, n)
        str2 = "damp = %20.14e    wantvar = %8g" % (damp, wantvar)
        str3 = "atol = %8.2e                 conlim = %8.2e" % (atol, conlim)
        str4 = "btol = %8.2e                 itnlim = %8g" % (btol, itnlim)
        print(str1)
        print(str2)
        print(str3)
        print(str4)

    itn = 0
    istop = 0
    nstop = 0
    ctol = 0
    if conlim > 0:
        ctol = 1.0 / conlim
    anorm = 0
    acond = 0
    dampsq = damp**2.0
    ddnorm = 0
    res2 = 0
    xnorm = 0
    xxnorm = 0
    z = 0
    cs2 = -1
    sn2 = 0

    # % Set up the first vectors u and v for the bidiagonalization.

    # % These satisfy  beta*u = b,  alfa*v = A'u.;'

    u = b[0:m]
    x = np.zeros(n)
    alfa = 0
    beta = np.linalg.norm(u)
    if beta > 0:
        u = u / beta
        v = Aprod(u, 2)
        nprodAT += 1
        alfa = np.linalg.norm(v)
    if alfa > 0:
        v = v / alfa
        w = v.copy()

    arnorm = alfa * beta
    if arnorm == 0:
        # if show, disp(msg(1,:)); end
        return
    arnorm0 = arnorm

    rhobar = alfa
    phibar = beta
    bnorm = beta
    rnorm = beta
    r1norm = rnorm
    r2norm = rnorm
    head1 = "   Itn      x(1)       r1norm     r2norm "
    head2 = " Compatible   LS      Norm A   Cond A"

    if show:
        print(" ")
        print(head1 + head2)
        test1 = 1
        test2 = alfa / beta
        str1 = "%6g %12.5e" % (itn, x[0])
        str2 = " %10.3e %10.3e" % (r1norm, r2norm)
        str3 = "  %8.1e %8.1e" % (test1, test2)
        print(str1 + str2 + str3)

    # %------------------------------------------------------------------
    # %     Main iteration loop.
    # %------------------------------------------------------------------
    while itn < itnlim:
        itn = itn + 1
        # %     Perform the next step of the bidiagonalization to obtain the
        # %     next  beta, u, alfa, v.  These satisfy the relations
        # %                beta*u  =  a*v   -  alfa*u,
        # %                alfa*v  =  A'*u  -  beta*v.'

        u = Aprod(v, 1) - alfa * u
        nprodA += 1
        beta = np.linalg.norm(u)
        if beta > 0:
            u = u / beta
            anorm = np.linalg.norm([anorm, alfa, beta, damp])
            v = Aprod(u, 2) - beta * v
            nprodAT += 1
            alfa = np.linalg.norm(v)
            if alfa > 0:
                v = v / alfa

        # %     Use a plane rotation to eliminate the damping parameter.
        # %     This alters the diagonal (rhobar) of the lower-bidiagonal matrix.

        rhobar1 = np.linalg.norm([rhobar, damp])
        cs1 = rhobar / rhobar1
        sn1 = damp / rhobar1
        psi = sn1 * phibar
        phibar = cs1 * phibar

        # %     Use a plane rotation to eliminate the subdiagonal element (beta)
        # %     of the lower-bidiagonal matrix, giving an upper-bidiagonal matrix.

        rho = np.linalg.norm([rhobar1, beta])
        cs = rhobar1 / rho
        sn = beta / rho
        theta = sn * alfa
        rhobar = -cs * alfa
        phi = cs * phibar
        phibar = sn * phibar
        tau = sn * phi

        # %     Update x and w.

        t1 = phi / rho
        t2 = -theta / rho
        dk = w / rho

        x = x + t1 * w
        w = v + t2 * w
        ddnorm = ddnorm + np.linalg.norm(dk) ** 2.0
        if wantvar:
            var = var + np.dot(dk, dk)

        # %     Use a plane rotation on the right to eliminate the
        # %     super-diagonal element (theta) of the upper-bidiagonal matrix.
        # %     Then use the result to estimate  norm(x).

        delta = sn2 * rho
        gambar = -cs2 * rho
        rhs = phi - delta * z
        zbar = rhs / gambar
        xnorm = np.sqrt(xxnorm + zbar**2.0)
        gamma = np.linalg.norm([gambar, theta])
        cs2 = gambar / gamma
        sn2 = theta / gamma
        z = rhs / gamma
        xxnorm = xxnorm + z**2.0

        # %     Test for convergence.
        # %     First, estimate the condition of the matrix  Abar,
        # %     and the norms of  rbar  and  Abar'rbar.'

        acond = anorm * np.sqrt(ddnorm)
        res1 = phibar**2.0
        res2 = res2 + psi**2.0
        rnorm = np.sqrt(res1 + res2)
        arnorm = alfa * abs(tau)

        # %     07 Aug 2002:
        # %     Distinguish between
        # %        r1norm = ||b - Ax|| and
        # %        r2norm = rnorm in current code
        # %               = sqrt(r1norm^2 + damp^2*||x||^2).
        # %        Estimate r1norm from
        # %        r1norm = sqrt(r2norm^2 - damp^2*||x||^2).
        # %     Although there is cancellation, it might be accurate enough.

        r1sq = rnorm**2.0 - dampsq * xxnorm
        r1norm = np.sqrt(abs(r1sq))
        if r1sq < 0:
            r1norm = -r1norm
        r2norm = rnorm.copy()

        # %     Now use these norms to estimate certain other quantities,
        # %     some of which will be small near a solution.

        test1 = rnorm / bnorm
        test2 = arnorm / arnorm0
        # %     test2   =   arnorm/( anorm * rnorm );
        test3 = 1.0 / acond
        t1 = test1 / (1.0 + anorm * xnorm / bnorm)
        rtol = btol + atol * anorm * xnorm / bnorm

        # %     The following tests guard against extremely small values of
        # %     atol, btol  or  ctol.  (The user may have set any or all of
        # %     the parameters  atol, btol, conlim  to 0.)
        # %     The effect is equivalent to the normal tests using
        # %     atol = eps,  btol = eps,  conlim = 1/eps.

        if itn >= itnlim:
            istop = 7
        if 1 + test3 <= 1:
            istop = 6
        if 1 + test2 <= 1:
            istop = 5
        if 1 + t1 <= 1:
            istop = 4

        # %     Allow for tolerances set by the user.

        if test3 <= ctol:
            istop = 3
        if test2 <= atol:
            istop = 2
        if test1 <= rtol:
            istop = 1

        # %     See if it is time to print something.
        if show:
            prnt = 0
            if n <= 40:
                prnt = 1
            if itn <= 10:
                prnt = 1
            if itn >= itnlim - 10:
                prnt = 1
            if itn % 10 == 0:
                prnt = 1
            if test3 <= 2 * ctol:
                prnt = 1
            if test2 <= 10 * atol:
                prnt = 1
            if test1 <= 10 * rtol:
                prnt = 1
            if istop != 0:
                prnt = 1

            if prnt == 1:
                str1 = "%6g %12.5e" % (itn, x[0])
                str2 = " %10.3e %10.3e" % (r1norm, r2norm)
                str3 = "  %8.1e %8.1e" % (test1, test2)
                str4 = " %8.1e %8.1e" % (anorm, acond)
                print(str1 + str2 + str3 + str4)
        if istop > 0:
            break

    # %     End of iteration loop.
    # %     Print the stopping condition.
    if show:
        print(" ")
        print("LSQR finished")
        print(msg[istop])
        print(" ")
        str1 = "istop =%8g   r1norm =%8.1e" % (istop, r1norm)
        str2 = "anorm =%8.1e   arnorm =%8.1e" % (anorm, arnorm)
        str3 = "itn   =%8g   r2norm =%8.1e" % (itn, r2norm)
        str4 = "acond =%8.1e   xnorm  =%8.1e" % (acond, xnorm)
        print(str1 + "   " + str2)
        print(str3 + "   " + str4)
        print(" ")
        # end

    print("nprodA", nprodA)
    print("nprodA", nprodAT)

    return x, istop, itn, r1norm, r2norm, anorm, acond, arnorm, xnorm, var



from .lsqr import lsqr
from .spgl1 import *


__all__ = [
    "oneprojector",
    "norm_l1nn_primal",
    "norm_l1nn_dual",
    "norm_l1nn_project",
    "norm_l12nn_primal",
    "norm_l12nn_dual",
    "norm_l12nn_project",
    "spgl1",
    "spg_bp",
    "spg_bpdn",
    "spg_lasso",
    "spg_mmv",
]


try:
    from .version import version as __version__
except ImportError:
    __version__ = "0.0.0"



r"""
MMV with non-negative norms
===========================
In this tutorial we will use the ``spg_mmv`` solver, which solves a
multi-measurement vector basis pursuit denoise problem. Both the standard L12
and L12non-negative norms will be compared. The latter set of norms is used if we
want our solution to have only positive values in case we have access to
such a kind of prior information.
"""
import matplotlib.pyplot as plt
import numpy as np

import spgl1

###############################################################################
# Let's first import our data, matrix operator and weights
data = np.load("../testdata/mmvnn.npz")

A = data["A"]
B = data["B"]
weights = data["weights"]

###############################################################################
# We apply unweighted inversions with and without non-negative
# norms.
X, _, _, info = spgl1.spg_mmv(A, B, 0.5, iter_lim=100, verbosity=0)

XNN, _, _, infoNN = spgl1.spg_mmv(
    A,
    B,
    0.5,
    iter_lim=100,
    verbosity=0,
    project=spgl1.norm_l12nn_project,
    primal_norm=spgl1.norm_l12nn_primal,
    dual_norm=spgl1.norm_l12nn_dual,
)
print("Negative X - MMV:", np.any(X < 0))
print("Negative X - MMVNN:", np.any(XNN < 0))
print("Residual norm - MMV:", info["rnorm"])
print("Residual norm - MMVNN:", infoNN["rnorm"])

plt.figure()
plt.plot(X[:, 0], "k", label="Coefficients")
plt.plot(XNN[:, 0], "--r", label="Coefficients NN")
plt.legend()
plt.title("Unweighted Basis Pursuit with Multiple Measurement Vectors")

plt.figure()
plt.plot(X[:, 1], "k", label="Coefficients")
plt.plot(XNN[:, 1], "--r", label="Coefficients NN")
plt.legend()
plt.title("Unweighted Basis Pursuit with Multiple Measurement Vectors")

###############################################################################
# We repeat the same steps with weighted norms.
X, _, _, info = spgl1.spg_mmv(A, B, 0.5, iter_lim=100, weights=np.array(weights), verbosity=0)
XNN, _, _, infoNN = spgl1.spg_mmv(
    A,
    B,
    0.5,
    iter_lim=100,
    verbosity=0,
    weights=np.array(weights),
    project=spgl1.norm_l12nn_project,
    primal_norm=spgl1.norm_l12nn_primal,
    dual_norm=spgl1.norm_l12nn_dual,
)
print("Negative X - MMV:", np.any(X < 0))
print("Negative X - MMVNN:", np.any(XNN < 0))
print("Residual norm - MMV:", info["rnorm"])
print("Residual norm - MMVNN:", infoNN["rnorm"])

plt.figure()
plt.plot(X[:, 0], "k", label="Coefficients")
plt.plot(XNN[:, 0], "--r", label="Coefficients NN")
plt.legend()
plt.title("Weighted Basis Pursuit with Multiple Measurement Vectors")

plt.figure()
plt.plot(X[:, 1], "k", label="Coefficients")
plt.plot(XNN[:, 1], "--r", label="Coefficients NN")
plt.legend()
plt.title("Weighted Basis Pursuit with Multiple Measurement Vectors")
